//#define LPC2148

#define false 0
#define true 1

enum {
  I2C_ASSERT_ACK = 0x04,
  I2C_INTERRUPT  = 0x08,
  I2C_STOP       = 0x10,
  I2C_START      = 0x20,
  I2C_ENABLE     = 0x40
};
 
static volatile uint8_t  _i2cSlave;
static volatile uint8_t* _i2cCommand;
static volatile int32_t  _i2cCommandLen;
static volatile int32_t  _i2cCommandPtr;
static volatile uint8_t* _i2cResponse;
static volatile int32_t  _i2cResponseLen;


/*
 *Configuration the clock rate
 */
void i2c0_clock_setup()
{
  I20SCLL = 0x1E; //30 for each one, sets I2C clock at 400Khz
  I20SCLH = 0x1E;
	
}

/* expiration timer for i2c transaction */
static void __attribute__ ((interrupt("IRQ"))) timer_interrupt(void) {
	vicDisable(INT_CHANNEL_TIMER1);
	vicDisable(INT_CHANNEL_I2C0);
	I20CONCLR = I2C_INTERRUPT | I2C_START | I2C_STOP;
	i2cResponsePtr = -2;
  	i2cDone = true;
	DBG("timer expired\n");
	vicUpdatePriority();
}

/* I2C timeout */
void enable_timer(){
	DBG("timer enabled\n");
	T1MR0 = 300000;
	T1MCR = BIT0 | BIT1;
	T1TCR = BIT0;	
	vicEnable(INT_CHANNEL_TIMER1);
}

void disable_timer(){
	vicDisable(INT_CHANNEL_TIMER1);
	T1MCR = 0;
	T1TCR = 0;
}
	
/*
 * A master transaction.
 * The master sends to the named slave the command,
 * and expects a response of a given length, which
 * it writes into the response buffer. It returns
 * the number of bytes received from the slave, or
 * a negative number in case of an error.
 */

//i2cMasterStransact -> now a non blocking function, only sets the initial interrupt
void i2cMasterTransact(uint8_t  slave,
                          uint8_t* command,
                          int32_t  command_len,
                          uint8_t* response,
                          int32_t  response_len)
{
  vicDisable(INT_CHANNEL_I2C0);

  i2cDone        = false;
  _i2cSlave       = slave;
  _i2cCommand     = command;
  _i2cCommandLen  = command_len;
  _i2cResponse    = response;
  _i2cResponseLen = response_len;
  _i2cCommandPtr  = 0;
  i2cResponsePtr = 0;
 
  I20CONSET = I2C_START;

  vicEnable(INT_CHANNEL_I2C0);
}
 
static void
  __attribute__ ((interrupt("IRQ")))
  _i2cIsr(void)
{
  disable_timer();
  int32_t status = I20STAT;
  DBG("_i2cCommandLen: ");
  DBGNUM(_i2cCommandLen);
  DBG("\n_i2cCommandPtr: ");
  DBGNUM(_i2cCommandPtr);
  DBG("\n");
  DBG("status: ");
  DBGNUM(status);
  DBG("\n");
  DBG("slave: ");
  DBGHEX(_i2cSlave, 2);
  DBG("\n");
 

  switch(status) {
      case 0x08: /* START has been transmitted            */
      case 0x10: /* repeated START has been transmitted   */
      if (_i2cCommandPtr == _i2cCommandLen)
        I20DAT = _i2cSlave | 0x01; /* read flag set       */
      else
        I20DAT = _i2cSlave; /* read flag not set, write   */
      I20CONCLR = I2C_INTERRUPT | I2C_STOP | I2C_START;
      break;

      case 0x18: /* slave address + write transmitted     */
      case 0x28: /* data has been transmitted in master   */
                 /* transmitter mode */
      if (_i2cCommandPtr < _i2cCommandLen) { /* send more */
        I20DAT = _i2cCommand[ _i2cCommandPtr ];
        _i2cCommandPtr++;
      I20CONCLR = I2C_INTERRUPT | I2C_START | I2C_STOP;
      } else { /* nothing more to send                    */
        if (_i2cResponseLen == 0) { /* no response req.   */
        I20CONSET = I2C_STOP;
        I20CONCLR = I2C_INTERRUPT | I2C_START;
        i2cDone = true;
        } else { /* need response, send repeated start    */
        I20CONSET = I2C_START;
        I20CONCLR = I2C_INTERRUPT | I2C_STOP;
        }
      }
      break;
     
      case 0x20: /* NOT ACK, master trans. after SLA+W    */
      case 0x48: /* NOT ACK, master recv. after SLA+R     */
      case 0x30: /* NOT ACK, master trans. after data     */
    I20CONSET = I2C_STOP;
    I20CONCLR = I2C_INTERRUPT | I2C_START;
    i2cDone = true;
    i2cResponsePtr = -status; /* this is an error      */
    break;       
     
      case 0x38: /* arbitration lost in master mode       */
    I20CONCLR = I2C_INTERRUPT | I2C_START | I2C_STOP;
    /* we mark this as an error and exit, although this */
    /* is not a fatal error; can try again.             */
    i2cDone = true;
    i2cResponsePtr = -status;
    break;
   
      case 0x40: /* ACK after SLA+R in master recv mode   */
      if (i2cResponsePtr < _i2cResponseLen-1)
      I20CONSET = I2C_ASSERT_ACK; /* we want more!      */
    else
      I20CONCLR = I2C_ASSERT_ACK; /* received enough    */         
    I20CONCLR = I2C_INTERRUPT | I2C_START | I2C_STOP;
    break;

      case 0x58: /* master received byte, sent NACK       */
      _i2cResponse[ i2cResponsePtr ] = I20DAT;
      i2cResponsePtr++;
    I20CONSET = I2C_STOP;
    I20CONCLR = I2C_INTERRUPT | I2C_START;
    i2cDone = true;
    break;
   
      case 0x50: /* master received byte, sent ACK        */
      _i2cResponse[ i2cResponsePtr ] = I20DAT;
      i2cResponsePtr++;
      if (i2cResponsePtr < _i2cResponseLen-1)
      I20CONSET = I2C_ASSERT_ACK; /* we want more       */
    else
      I20CONCLR = I2C_ASSERT_ACK; /* received enough    */         
    I20CONCLR = I2C_INTERRUPT | I2C_START | I2C_STOP;
    break;       
  
    default:
    i2cDone = true;
    i2cResponsePtr = -status;
    break;   
  }

  if(!i2cDone) //I2C interrrupt chain not finished, set timeout again
	enable_timer();
  vicUpdatePriority();
}

/* only master mode */
void i2c_init()
{
  uint8_t i2c0_isr_priority = 0;
  i2c0_clock_setup();
  PINSEL0      &= ~0xF0; /* P0.3 = SDA, P0.2 = SCL */
  PINSEL0      |= 0x50;
  //I20ADR       = slave_address;
  //I20CONSET    = I2C_ENABLE | I2C_ASSERT_ACK;
  I20CONSET = I2C_ENABLE;
  I20CONCLR = BIT2 | BIT3 | BIT4 | BIT5; 
 
  vicEnableIRQ(INT_CHANNEL_I2C0,
               i2c0_isr_priority, /* priority */
               _i2cIsr);
  vicEnableIRQ(INT_CHANNEL_TIMER1, 2, timer_interrupt);
  vicDisable(INT_CHANNEL_TIMER1);
  vicUpdatePriority();
}


int getBufferI2C(int8_t *buffer)
{
	int i;
	//getting next 3 bytes of the protocol - protocol, add, write_num, read_num
	for (i=0; i < 3; i++)
	{
		buffer[i] = vcomGetchar();
		if (buffer[i] == -1)//not enough data from user, send error
			return -1;

		DBG("buffer: ");
		DBGNUM(i);
		DBG(" ");
		DBGNUM(buffer[i]);
		DBG("\n");
		
	}
	//getting the write data
	for (i=3;i < 3 + buffer[1]; i++)
	{
		buffer[i] = vcomGetchar();
		if (buffer[i] == -1)//not enough data from user, send error
			return -1;
		
		DBG("buffer: ");
		DBGNUM(i);
		DBG(" ");
		DBGNUM(buffer[i]);
		DBG("\n");
	}
	return 0;
}

void handleI2C(int8_t *ins_buffer)
{
	//copy write to i2c_write				
	memcpy(i2c_write_buffer, ins_buffer + 3, ins_buffer[1]);
	DBG("calling i2cMasterTransact for request: ");
	DBGNUM(i2c_busy);
	DBG("\n");
	i2cMasterTransact(ins_buffer[0], i2c_write_buffer, ins_buffer[1], i2c_read_buffer, ins_buffer[2]);
}



