/*
	
	Copyright (C) 2006 Bertrik Sikken (bertrik@sikken.nl)

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	1. Redistributions of source code must retain the above copyright
	   notice, this list of conditions and the following disclaimer.
	2. Redistributions in binary form must reproduce the above copyright
	   notice, this list of conditions and the following disclaimer in the
	   documentation and/or other materials provided with the distribution.
	3. The name of the author may not be used to endorse or promote products
	   derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
	IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
	OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
	IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, 
	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
	THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
	THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
	Minimal implementation of a USB serial port, using the CDC class.
	This example application simply echoes everything it receives right back
	to the host.

	Windows:
	Extract the usbser.sys file from .cab file in C:\WINDOWS\Driver Cache\i386
	and store it somewhere (C:\temp is a good place) along with the usbser.inf
	file. Then plug in the LPC214x and direct windows to the usbser driver.
	Windows then creates an extra COMx port that you can open in a terminal
	program, like hyperterminal.

	Linux:
	The device should be recognised automatically by the cdc_acm driver,
	which creates a /dev/ttyACMx device file that acts just like a regular
	serial port.
  
  Modified by Sivan Toledo and Wouter van Ooijen

*/


#include <stdint.h>
#include <string.h> /* for memcpy */

//public variables for main loop to chak and i2c to set
static volatile int32_t  i2cResponsePtr;
static volatile int32_t  i2cDone;


#include <lpc2000/io.h>
#include <lpc2000/interrupt.h>
#include "debug/debug.h"
#include "lpc2000/vic.c"
#include "protocols.h"

typedef int bool; /* used by usb.h, usb-misc.c */

#include "usb/usb.h"
#include "usb/fifo.c"
#include "usb/usb_misc.c"

#define usb_SOF_handler() /* used in usb.c */
#include "usb/usb.c"

#define MAX_RW_LENGTH 128
#define MAX_MSG_LENGTH (MAX_RW_LENGTH + 4)

//protocol vars and buffers
static uint8_t protocol; //i2c, uart...
static uint8_t protocol_addr; //lm75, lcd
static uint8_t operation; //read, write
static uint8_t numBytes;
static uint8_t data_buffer[128];
static uint8_t data_ptr = 0;

int request_num = 1; //first request will be 1

int i2c_busy = 0; //flag for i2c busy state
int8_t user_buffer[MAX_MSG_LENGTH]; //buffer to hold user incoming message
uint8_t i2c_write_buffer[MAX_RW_LENGTH]; //buffer to hold trasmitted data to i2c
uint8_t i2c_read_buffer[MAX_RW_LENGTH]; //buffer to hold received data from i2c
uint32_t gpio_ret; //hold the IOPIN
int handle_new_req = 0; //flag to signal arrival of new request from user

//our protocols implementations
#include "i2c/i2c.c"
#include "gpio/gpio.c"

/*
Init function for all of the serial bridge components
*/
void init(){
  int i;

  uart0Init(); //for DEBUG
  DBG("start prog\n");
 
  //Setting clock
  VPBDIV = 0x00000001;  /* PCLK = CCLK */
  PLLCFG  = 0x00000021; /* Fosc=12MHz, CCLK=24MHz */
  PLLCON  = 0x00000001; /* enable the PLL  */
  PLLFEED = 0x000000AA; /* feed sequence   */
  PLLFEED = 0x00000055;
  while (!(PLLSTAT & 0x00000400));
  PLLCON = 3;     // enable and connect
  PLLFEED = 0xAA;
  PLLFEED = 0x55;

  DBG("after clk\n");
  interruptsEnable(); 
  vicInit();
  DBG("Initializing USB stack\n");
  usbInit();
  i2c_init();
  DBG("Starting USB communication\n");
  usbConnect();

  DBG("before timer\n");
  //waiting for usb to finish identifing to OS - clear buffers
  for (i=0; i< 150; i++){
    T0TC = 0;
    T0MR0 = 2999999;
    T0MCR = BIT2;
    T0TCR = BIT0;
    while (T0TCR & BIT0);
  }
  DBG("after timer\n");
  vcomInit();
}

void sendError()
{
	vcomPutchar(PROTOCOL_BUSY);
	DBG("send Error\n");
}



int main(void) {
	int i;
	int rv;
	uint8_t one_byte;
	
	init();

	while(1){
		//check if user send reqeust
		user_buffer[0] = vcomGetchar();
		if (user_buffer[0] != -1) //got user msg
		{
			//read messages, mark messages to handle, send ack/nack, handle immediate protocols
			switch(user_buffer[0])
			{				
				case PROTOCOL_I2C:
					//get buffers
					rv = getBufferI2C(user_buffer + 1);
					if (rv < 0)//failed to read i2c message
					{
						sendError();
						break;
					}
					DBG("in i2c\n");
					if (i2c_busy == 0)
					{
						handle_new_req = 1;
						i2c_busy = request_num;
						vcomPutchar(request_num++);//send ack
						if (request_num == 0xff)//request num are modulo 0xff
							request_num = 1;
						DBG("req num: ");
						DBGNUM(request_num);
						DBG("\n");
					}
					else
					{
						//send nack, i2c protocol busy
						sendError();
					}
					break;

				case PROTOCOL_GPIO:
					//get buffers					
					rv = getBufferGPIO(user_buffer + 1);
					if (rv < 0)//failed to read gpio message
					{
						sendError();
						break;
					}
					rv = handleGPIO(user_buffer + 1, &gpio_ret);
					if (rv < 0)//gpio internal error
						sendError();
					else
					{
						vcomPutchar(request_num++);
						if (request_num == 0xff)//request num are modulo 0xff
							request_num = 1;
						DBG("req num: ");
						DBGNUM(request_num);
						DBG("\n");
						if (rv == 1)//this marks a requested gpio value read
						{
							for (i=0; i<4; i++)//send response, little endian
							{							
								one_byte = (uint8_t) gpio_ret;
								DBG("sending ");
								DBGNUM(i);
								DBG(" ");
								DBGNUM(one_byte);
								DBG("\n");
								vcomPutchar(one_byte);
								gpio_ret>>=8;
							}	
						}	
					}
					break;
				default:
					break;
			}
		}

		//handle the reqeusts for non-blocking protocols
		if (handle_new_req == 1)
		{
			switch(user_buffer[0])
			{
				case PROTOCOL_I2C:
					handleI2C(user_buffer + 1);
					break;
				default:
					break;
			}
			handle_new_req = 0; //mark request handled
		}

		//check done states
		if (i2cDone && (i2c_busy != 0))
		{
			DBG("i2cDone for i2c_busy: ");
			DBGNUM(i2c_busy);
			DBG("\n");
			vcomPutchar(i2c_busy);
			if (i2cResponsePtr < 0)//error from i2c - replying error
			{				
				DBG("sending error: ");
				DBGNUM(i2cResponsePtr);
				DBG("\n");
				vcomPutchar(STATE_ERROR);
			}
			else //send i2c response to user
			{
				vcomPutchar(i2cResponsePtr);
				for (i=0; i < i2cResponsePtr ; i++)
					vcomPutchar(i2c_read_buffer[i]);				
			}
			i2c_busy = 0; //mark i2c not busy
		}
	}
}
